"""Module for search results."""

from __future__ import annotations
from dataclasses import dataclass
import os.path
from typing import Optional


@dataclass
class Result:
    """This models a search result.

    :param ident: The identifier of the entry in the source
    :type ident: str
    :param source: The name of the source of the entry
    :type source: str
    :param title: The title of the song
    :type title: str
    :param artist: The artist of the song
    :type artist: str
    :param album: The name of the album or compilation, this particular
        version is from.
    :type album: str
    """

    ident: str
    source: str
    title: str
    artist: str
    album: str
    duration: Optional[str] = None

    @classmethod
    def from_filename(cls, filename: str, source: str) -> Result:
        """
        Infer most attributes from the filename.

        The filename must be in this form::

            {artist} - {title} - {album}.ext

        If parsing failes, the filename will be used as the title and the
        artist and album will be set to "Unknown".

        :param filename: The filename to parse
        :type filename: str
        :param source: The name of the source
        :type source: str
        :return: see above
        :rtype: Optional[Result]
        """
        basename = os.path.splitext(filename)[0]
        try:
            splitfile = os.path.basename(basename).split(" - ")
            ident = filename
            artist = splitfile[0].strip()
            title = splitfile[1].strip()
            album = splitfile[2].strip()
            return cls(ident=ident, source=source, title=title, artist=artist, album=album)
        except IndexError:
            return cls(
                ident=filename, source=source, title=basename, artist="Unknown", album="Unknown"
            )

    @classmethod
    def from_dict(cls, values: dict[str, str]) -> Result:
        """
        Create a Result object from a dictionary.

        The dictionary must have the following keys:
          - ident (str)
          - source (str)
          - title (str)
          - artist (str)
          - album (str)
          - duration (int, optional)

        :param values: The dictionary with the values
        :type values: dict[str, str]
        :return: The Result object
        :rtype: Result
        """
        return cls(
            ident=values["ident"],
            source=values["source"],
            title=values["title"],
            artist=values["artist"],
            album=values["album"],
            duration=values.get("duration", None),
        )

    def to_dict(self) -> dict[str, str]:
        """
        Convert the Result object to a dictionary.

        The dictionary will have the following keys:
          - ident (str)
          - source (str)
          - title (str)
          - artist (str)
          - album (str)
          - duration (str, if available)

        :return: The dictionary with the values
        :rtype: dict[str, str]
        """
        output = {
            "ident": self.ident,
            "source": self.source,
            "title": self.title,
            "artist": self.artist,
            "album": self.album,
        }
        if self.duration is not None:
            output["duration"] = self.duration
        return output



"""
Module for the Server.

The server listens for incoming connections from playback clients and web
clients via the socket.io protocol.

It manages multiple independent rooms, each with its own queue and configuration.
If configured, the server can be in private mode, where only playback clients with
a valid registration key can connect. It can also be in restricted mode, where only
search is forwarded to the playback client, unless the client has a valid registration
key.
"""

from __future__ import annotations

import asyncio
import datetime
import hashlib
import logging
import os
import random
import string
from json.decoder import JSONDecodeError
from argparse import Namespace
from dataclasses import dataclass
from dataclasses import field
from typing import Any
from typing import AsyncGenerator
from typing import Optional

import socketio
from aiohttp import web
from profanity_check import predict

from .result import Result
from .sources.youtube import YouTube

from . import jsonencoder
from .entry import Entry
from .queue import Queue
from .sources import available_sources
from .sources import Source

sio = socketio.AsyncServer(
    cors_allowed_origins="*", logger=True, engineio_logger=False, json=jsonencoder
)
app = web.Application()
sio.attach(app)

DEFAULT_CONFIG = {
    "preview_duration": 3,
    "waiting_room_policy": None,
    "last_song": None,
}


async def root_handler(request: Any) -> Any:
    """
    Handle the index and favicon requests.

    If the path of the request ends with "/favicon.ico" return the favicon,
    otherwise the index.html. This way the javascript can read the room code
    from the url.

    :param request Any: Webrequest from aiohttp
    :return: Either the favicon or the index.html
    :rtype web.FileResponse:
    """
    if request.path.endswith("/favicon.ico"):
        return web.FileResponse(os.path.join(app["root_folder"], "favicon.ico"))
    return web.FileResponse(os.path.join(app["root_folder"], "index.html"))


logging.basicConfig(level=logging.WARNING)
logger = logging.getLogger(__name__)


@dataclass
class Client:
    """This stores the configuration of a specific playback client.

    In case a new playback client connects to a room, these values can be
    overwritten.

    :param sources: A dictionary mapping the name of the used sources to their
        instances.
    :type sources: Source
    :param sources_prio: A list defining the order of the search results.
    :type sources_prio: list[str]
    :param config: Various configuration options for the client:
        * `preview_duration` (`Optional[int]`): The duration in seconds the
            playback client shows a preview for the next song. This is accounted for
            in the calculation of the ETA for songs later in the queue.
        * `last_song` (`Optional[float]`): A timestamp, defining the end of the queue.
        * `waiting_room_policy` (Optional[str]): One of:
            - `forced`, if a performer is already in the queue, they are put in the
                       waiting room.
            - `optional`, if a performer is already in the queue, they have the option
                          to be put in the waiting room.
            - `None`, performers are always added to the queue.
    :type config: dict[str, Any]:
    """

    sources: dict[str, Source]
    sources_prio: list[str]
    config: dict[str, Any]


@dataclass
class State:
    """This defines the state of one session/room.

    :param secret: The secret for the room. Used to log in as an admin on the
        webclient or reconnect a playbackclient
    :type secret: str
    :param queue: A queue of :py:class:`syng.entry.Entry` objects. New songs
        are appended to this, and if a playback client requests a song, it is
        taken from the top.
    :type queue: Queue
    :param waiting_room: Contains the Entries, that are hold back, until a
        specific song is finished.
    :type waiting_room: list[Entry]
    :param recent: A list of already played songs in order.
    :type recent: list[Entry]
    :param sid: The socket.io session id of the (unique) playback client. Once
        a new playback client connects to a room (with the correct secret),
        this will be swapped with the new sid.
    :type sid: str
    :param client: The config for the playback client
    :type client: Client
    :param last_seen: Timestamp of the last connected client. Used to determine
        if a room is still in use.
    :type last_seen: datetime
    """

    queue: Queue
    waiting_room: list[Entry]
    recent: list[Entry]
    sid: str
    client: Client
    last_seen: datetime.datetime = field(init=False, default_factory=datetime.datetime.now)


clients: dict[str, State] = {}


async def send_state(state: State, sid: str) -> None:
    """
    Send the current state (queue and recent-list) to sid.

    This sends a "state" message. This can be received either by the playback
    client, a web client or the whole room.

    If it is send to a playback client, it will be handled by the
    :py:func:`syng.client.handle_state` function.

    :param state: The state to send
    :type state: State
    :param sid: The recepient of the "state" message
    :type sid: str:
    :rtype: None
    """

    safe_config = {k: v for k, v in state.client.config.items() if k not in ["secret", "key"]}

    await sio.emit(
        "state",
        {
            "queue": state.queue,
            "recent": state.recent,
            "waiting_room": state.waiting_room,
            "config": safe_config,
        },
        room=sid,
    )


@sio.on("get-state")
async def handle_state(sid: str) -> None:
    """
    Handle the "get-state" message.

    Sends the current state to whoever requests it. This failes if the sender
    is not part of any room.

    :param sid: The initial sender, and therefore recepient of the "state"
        message
    :type sid: str
    :rtype: None
    """
    async with sio.session(sid) as session:
        room = session["room"]
    state = clients[room]

    await send_state(state, sid)


@sio.on("waiting-room-append")
async def handle_waiting_room_append(sid: str, data: dict[str, Any]) -> None:
    """
    Append a song to the waiting room.

    This should be called from a web client. Appends the entry, that is encoded
    within the data to the waiting room of the room the client is currently
    connected to.

    :param sid: The session id of the client sending this request
    :type sid: str
    :param data: A dictionary encoding the entry, that should be added to the
        waiting room.
    :type data: dict[str, Any]
    :rtype: None
    """
    async with sio.session(sid) as session:
        room = session["room"]
    state = clients[room]

    source_obj = state.client.sources[data["source"]]

    entry = await source_obj.get_entry(data["performer"], data["ident"])

    if entry is None:
        await sio.emit(
            "msg",
            {"msg": f"Unable to add to the waiting room: {data['ident']}. Maybe try again?"},
            room=sid,
        )
        return

    if "uid" not in data or (
        (data["uid"] is not None and len(list(state.queue.find_by_uid(data["uid"]))) == 0)
        or (data["uid"] is None and state.queue.find_by_name(data["performer"]) is None)
    ):
        await append_to_queue(room, entry, sid)
        return

    entry.uid = data["uid"]

    state.waiting_room.append(entry)
    await send_state(state, room)
    await sio.emit(
        "get-meta-info",
        entry,
        room=clients[room].sid,
    )


async def append_to_queue(room: str, entry: Entry, report_to: Optional[str] = None) -> None:
    """
    Append a song to the queue for a given session.

    Checks, if the computed start time is before the configured end time of the
    event, and reports an error, if the end time is exceeded.

    :param room: The room with the queue.
    :type room: str
    :param entry: The entry that contains the song.
    :type entry: Entry
    :param report_to: If an error occurs, who to report to.
    :type report_to: Optional[str]
    :rtype: None
    """
    state = clients[room]

    first_song = state.queue.try_peek()
    if first_song is None or first_song.started_at is None:
        start_time = datetime.datetime.now().timestamp()
    else:
        start_time = first_song.started_at

    start_time = state.queue.fold(
        lambda item, time: time + item.duration + state.client.config["preview_duration"] + 1,
        start_time,
    )

    if state.client.config["last_song"]:
        if state.client.config["last_song"] < start_time:
            # end_time = datetime.datetime.fromtimestamp(
            #     state.client.config["last_song"]
            # )
            if report_to is not None:
                await sio.emit(
                    "err",
                    {
                        "type": "QUEUE_FULL",
                        "end_time": state.client.config["last_song"],
                    },
                    room=report_to,
                )
            return

    state.queue.append(entry)
    await send_state(state, room)

    await sio.emit(
        "get-meta-info",
        entry,
        room=clients[room].sid,
    )


@sio.on("show_config")
async def handle_show_config(sid: str) -> None:
    """
    Sends public config to webclient.

    This will only be send if the client is on an admin connection.

    :param sid: The session id of the client sending this request
    :type sid: str
    :rtype: None
    """

    async with sio.session(sid) as session:
        room = session["room"]
        is_admin = session["admin"]
    state = clients[room]

    if is_admin:
        await sio.emit(
            "config",
            state.client.config,
            sid,
        )
    else:
        await sio.emit("err", {"type": "NO_ADMIN"}, sid)


@sio.on("update_config")
async def handle_update_config(sid: str, data: dict[str, Any]) -> None:
    """
    Forwards an updated config from an authorized webclient to the playback client.

    This is currently untrested and should be used with caution.

    :param sid: The session id of the client sending this request
    :type sid: str
    :param data: A dictionary encoding the new configuration
    :type data: dict[str, Any]
    :rtype: None
    """
    async with sio.session(sid) as session:
        room = session["room"]
        is_admin = session["admin"]
    state = clients[room]

    if is_admin:
        try:
            config = jsonencoder.loads(data["config"])
            await sio.emit(
                "update_config",
                DEFAULT_CONFIG | config,
                state.sid,
            )
            state.client.config = DEFAULT_CONFIG | config
            await sio.emit("update_config", config, room)
        except JSONDecodeError:
            await sio.emit("err", {"type": "JSON_MALFORMED"})

    else:
        await sio.emit("err", {"type": "NO_ADMIN"}, sid)


@sio.on("append")
async def handle_append(sid: str, data: dict[str, Any]) -> None:
    """
    Handle the "append" message.

    This should be called from a web client. Appends the entry, that is encoded
    within the data to the room the client is currently connected to. An entry
    constructed this way, will be given a UUID, to differentiate it from other
    entries for the same song. Additionally an id of the web client is saved
    for that entry.

    If the room is configured to no longer accept songs past a certain time
    (via the :py:attr:`Config.last_song` attribute), it is checked, if the
    start time of the song would exceed this time. If this is the case, the
    request is denied and a "msg" message is send to the client, detailing
    this.

    If a waitingroom is forced or optional, it is checked, if one of the performers is
    already in queue. In that case, a "ask_for_waitingroom" message is send to the
    client.

    Otherwise the song is added to the queue. And all connected clients (web
    and playback client) are informed of the new state with a "state" message.

    Since some properties of a song can only be accessed on the playback
    client, a "get-meta-info" message is send to the playback client. This is
    handled there with the :py:func:`syng.client.handle_get_meta_info`
    function.

    :param sid: The session id of the client sending this request
    :type sid: str
    :param data: A dictionary encoding the entry, that should be added to the
        queue.
    :type data: dict[str, Any]
    :rtype: None
    """
    async with sio.session(sid) as session:
        room = session["room"]
    state = clients[room]

    if len(data["performer"]) > 50:
        await sio.emit("err", {"type": "NAME_LENGTH", "name": data["performer"]}, room=sid)
        return

    if predict([data["performer"]]) == [1]:
        await sio.emit("err", {"type": "PROFANITY", "name": data["performer"]}, room=sid)
        return

    if state.client.config["waiting_room_policy"] and (
        state.client.config["waiting_room_policy"].lower() == "forced"
        or state.client.config["waiting_room_policy"].lower() == "optional"
    ):
        old_entry = state.queue.find_by_name(data["performer"])
        if old_entry is not None:
            await sio.emit(
                "ask_for_waitingroom",
                {
                    "current_entry": {
                        "source": data["source"],
                        "performer": data["performer"],
                        "ident": data["ident"],
                    },
                    "old_entry": {
                        "artist": old_entry.artist,
                        "title": old_entry.title,
                        "performer": old_entry.performer,
                    },
                },
                room=sid,
            )
            return

    source_obj = state.client.sources[data["source"]]

    entry = await source_obj.get_entry(data["performer"], data["ident"])

    if entry is None:
        await sio.emit(
            "msg",
            {"msg": f"Unable to append {data['ident']}. Maybe try again?"},
            room=sid,
        )
        return

    entry.uid = data["uid"] if "uid" in data else None

    await append_to_queue(room, entry, sid)


@sio.on("append-anyway")
async def handle_append_anyway(sid: str, data: dict[str, Any]) -> None:
    """
    Appends a song to the queue, even if the performer is already in queue.

    Works the same as handle_append, but without the check if the performer is already
    in queue.

    Only if the waiting_room_policy is not configured as forced.
    """
    async with sio.session(sid) as session:
        room = session["room"]
    state = clients[room]

    if len(data["performer"]) > 50:
        await sio.emit("err", {"type": "NAME_LENGTH", "name": data["performer"]}, room=sid)
        return

    if predict([data["performer"]]) == [1]:
        await sio.emit("err", {"type": "PROFANITY", "name": data["performer"]}, room=sid)
        return

    if state.client.config["waiting_room_policy"].lower() == "forced":
        await sio.emit(
            "err",
            {"type": "WAITING_ROOM_FORCED"},
            room=sid,
        )
        return

    source_obj = state.client.sources[data["source"]]

    entry = await source_obj.get_entry(data["performer"], data["ident"])

    if entry is None:
        await sio.emit(
            "msg",
            {"msg": f"Unable to append {data['ident']}. Maybe try again?"},
            room=sid,
        )
        return

    entry.uid = data["uid"] if "uid" in data else None

    await append_to_queue(room, entry, sid)


@sio.on("meta-info")
async def handle_meta_info(sid: str, data: dict[str, Any]) -> None:
    """
    Handle the "meta-info" message.

    Updated a :py:class:syng.entry.Entry`, that is encoded in the data
    parameter, in the queue, that belongs to the room the requesting client
    belongs to, with new meta data, that is send from the playback client.

    Afterwards send the updated queue to all members of the room.

    :param sid: The session id of the client sending this request.
    :type sid: str
    :param data: A dictionary encoding the entry to update (already with the
        new metadata)
    :type data: dict[str, Any]
    :rtype: None
    """
    async with sio.session(sid) as session:
        room = session["room"]
    state = clients[room]

    state.queue.update(
        data["uuid"],
        lambda item: item.update(**data["meta"]),
    )

    for entry in state.waiting_room:
        if entry.uuid == data["uuid"] or str(entry.uuid) == data["uuid"]:
            entry.update(**data["meta"])

    await send_state(state, room)


@sio.on("get-first")
async def handle_get_first(sid: str) -> None:
    """
    Handle the "get-first" message.

    This message is send by the playback client, once it has connected. It
    should only be send for the initial song. Each subsequent song should be
    requestet with a "pop-then-get-next" message (See
    :py:func:`handle_pop_then_get_next`).

    If no songs are in the queue for this room, this function waits until one
    is available, then notes its starting time and sends it back to the
    playback client in a "play" message. This will be handled by the
    :py:func:`syng.client.handle_play` function.

    :param sid: The session id of the requesting client
    :type sid: str
    :rtype: None
    """
    async with sio.session(sid) as session:
        room = session["room"]
    state = clients[room]

    current = await state.queue.peek()
    current.started_at = datetime.datetime.now().timestamp()

    await sio.emit("play", current, room=sid)


@sio.on("waiting-room-to-queue")
async def handle_waiting_room_to_queue(sid: str, data: dict[str, Any]) -> None:
    """
    Handle the "waiting-room-to-queue" message.

    If on an admin-connection, removes a song from the waiting room and appends it to
    the queue.

    :param sid: The session id of the requesting client
    :type sid: str
    :rtype: None
    """
    async with sio.session(sid) as session:
        room = session["room"]
        is_admin = session["admin"]
    state = clients[room]

    if is_admin:
        entry = next(
            (wr_entry for wr_entry in state.waiting_room if str(wr_entry.uuid) == data["uuid"]),
            None,
        )
        if entry is not None:
            state.waiting_room.remove(entry)
            await append_to_queue(room, entry, sid)


async def add_songs_from_waiting_room(room: str) -> None:
    """
    Add all songs from the waiting room, that should be added to the queue.

    A song should be added if none of its performers are already queued.

    This should be called every time a song leaves the queue.

    :param room: The room holding the queue.
    :type room: str
    :rtype: None
    """
    state = clients[room]

    wrs_to_remove = []
    for wr_entry in state.waiting_room:
        if state.queue.find_by_name(wr_entry.performer) is None:
            await append_to_queue(room, wr_entry)
            wrs_to_remove.append(wr_entry)

    for wr_entry in wrs_to_remove:
        state.waiting_room.remove(wr_entry)


async def discard_first(room: str) -> Entry:
    """
    Gets the first element of the queue, handling resulting triggers.

    This function is used to get the first element of the queue, and handle
    the resulting triggers. This includes adding songs from the waiting room,
    and updating the state of the room.

    :param room: The room to get the first element from.
    :type room: str
    :rtype: Entry
    """
    state = clients[room]

    old_entry = await state.queue.popleft()

    await add_songs_from_waiting_room(room)

    state.recent.append(old_entry)
    state.last_seen = datetime.datetime.now()

    return old_entry


@sio.on("pop-then-get-next")
async def handle_pop_then_get_next(sid: str) -> None:
    """
    Handle the "pop-then-get-next" message.

    This function acts similar to the :py:func:`handle_get_first` function. The
    main difference is, that prior to sending a song to the playback client,
    the first element of the queue is discarded.

    Afterwards it follows the same steps as the handler for the "play" message,
    get the first element of the queue, annotate it with the current time,
    update everyones state and send the entry it to the playback client in a
    "play" message. This will be handled by the
    :py:func:`syng.client.handle_play` function.

    :param sid: The session id of the requesting playback client
    :type sid: str
    :rtype: None
    """
    async with sio.session(sid) as session:
        room = session["room"]
    state = clients[room]

    if sid != state.sid:
        return

    await discard_first(room)
    await send_state(state, room)

    current = await state.queue.peek()
    current.started_at = datetime.datetime.now().timestamp()
    await send_state(state, room)

    await sio.emit("play", current, room=sid)


def check_registration(key: str) -> bool:
    """
    Check if a given key is in the registration keyfile.

    This is used to authenticate a client, if the server is in private or
    restricted mode.

    :param key: The key to check
    :type key: str
    :return: True if the key is in the registration keyfile, False otherwise
    :rtype: bool
    """
    with open(app["registration-keyfile"], encoding="utf8") as f:
        raw_keys = f.readlines()
        keys = [key[:64] for key in raw_keys]

        return key in keys


@sio.on("register-client")
async def handle_register_client(sid: str, data: dict[str, Any]) -> None:
    """
    Handle the "register-client" message.

    The data dictionary should have the following keys:
        - `room` (Optional), the requested room
        - `config`, an dictionary of initial configurations
        - `queue`, a list of initial entries for the queue. The entries are
                    encoded as a dictionary.
        - `recent`, a list of initial entries for the recent list. The entries
                    are encoded as a dictionary.
        - `secret`, the secret of the room
        - `key`, a registration key given out by the server administrator

    This will register a new playback client to a specific room. If there
    already exists a playback client registered for this room, this
    playback client will be replaced if and only if, the new playback
    client has the same secret.

    If registration is restricted, abort, if the given key is not in the
    registration keyfile.

    If no room is provided, a fresh room id is generated.

    If the client provides a new room, or a new room id was generated, the
    server will create a new :py:class:`State` object and associate it with
    the room id. The state will be initialized with a queue and recent
    list, an initial config as well as no sources (yet).

    In any case, the client will be notified of the success or failure, along
    with its assigned room key via a "client-registered" message. This will be
    handled by the :py:func:`syng.client.handle_client_registered` function.

    If it was successfully registerd, the client will be added to its assigend
    or requested room.

    Afterwards all clients in the room will be send the current state.

    :param sid: The session id of the requesting playback client.
    :type sid: str
    :param data: A dictionary with the keys described above
    :type data: dict[str, Any]
    :rtype: None
    """

    def gen_id(length: int = 4) -> str:
        client_id = "".join([random.choice(string.ascii_letters) for _ in range(length)])
        if client_id in clients:
            client_id = gen_id(length + 1)
        return client_id

    if "key" in data["config"]:
        data["config"]["key"] = hashlib.sha256(data["config"]["key"].encode()).hexdigest()

    if app["type"] == "private" and (
        "key" not in data["config"] or not check_registration(data["config"]["key"])
    ):
        await sio.emit(
            "client-registered",
            {"success": False, "room": None},
            room=sid,
        )
        return

    room: str = (
        data["config"]["room"] if "room" in data["config"] and data["config"]["room"] else gen_id()
    )
    async with sio.session(sid) as session:
        session["room"] = room

    if room in clients:
        old_state: State = clients[room]
        if data["config"]["secret"] == old_state.client.config["secret"]:
            logger.info("Got new client connection for %s", room)
            old_state.sid = sid
            old_state.client = Client(
                sources=old_state.client.sources,
                sources_prio=old_state.client.sources_prio,
                config=DEFAULT_CONFIG | data["config"],
            )
            await sio.enter_room(sid, room)
            await sio.emit("client-registered", {"success": True, "room": room}, room=sid)
            await send_state(clients[room], sid)
        else:
            logger.warning("Got wrong secret for %s", room)
            await sio.emit("client-registered", {"success": False, "room": room}, room=sid)
    else:
        logger.info("Registerd new client %s", room)
        initial_entries = [Entry(**entry) for entry in data["queue"]]
        initial_waiting_room = [Entry(**entry) for entry in data["waiting_room"]]
        initial_recent = [Entry(**entry) for entry in data["recent"]]

        clients[room] = State(
            queue=Queue(initial_entries),
            waiting_room=initial_waiting_room,
            recent=initial_recent,
            sid=sid,
            client=Client(
                sources={},
                sources_prio=[],
                config=DEFAULT_CONFIG | data["config"],
            ),
        )

        await sio.enter_room(sid, room)
        await sio.emit("client-registered", {"success": True, "room": room}, room=sid)
        await send_state(clients[room], sid)


@sio.on("sources")
async def handle_sources(sid: str, data: dict[str, Any]) -> None:
    """
    Handle the "sources" message.

    Get the list of sources the client wants to use. Update internal list of
    sources, remove unused sources and query for a config for all uninitialized
    sources by sending a "request-config" message for each such source to the
    playback client. This will be handled by the
    :py:func:`syng.client.request-config` function.

    This will not yet add the sources to the configuration, rather gather what
    sources need to be configured and request their configuration. The list
    of sources will set the :py:attr:`Config.sources_prio` attribute.

    :param sid: The session id of the playback client
    :type sid: str
    :param data: A dictionary containing a "sources" key, with the list of
        sources to use.
    :type data: dict[str, Any]
    :rtype: None
    """
    async with sio.session(sid) as session:
        room = session["room"]
    state = clients[room]

    if sid != state.sid:
        return

    unused_sources = state.client.sources.keys() - data["sources"]
    new_sources = data["sources"] - state.client.sources.keys()

    for source in unused_sources:
        del state.client.sources[source]

    state.client.sources_prio = data["sources"]

    for name in new_sources:
        await sio.emit("request-config", {"source": name}, room=sid)


@sio.on("config-chunk")
async def handle_config_chunk(sid: str, data: dict[str, Any]) -> None:
    """
    Handle the "config-chunk" message.

    This is called, when a source wants its configuration transmitted in
    chunks, rather than a single message. If the source already exist
    (e.g. when this is not the first chunk), the config will be added
    to the source, otherwise a source will be created with the given
    configuration.

    :param sid: The session id of the playback client
    :type sid: str
    :param data: A dictionary with a "source" (str) and a
        "config" (dict[str, Any]) entry. The exact content of the config entry
        depends on the source.
    :rtype: None
    """
    async with sio.session(sid) as session:
        room = session["room"]
    state = clients[room]

    if sid != state.sid:
        return

    if data["source"] not in state.client.sources:
        state.client.sources[data["source"]] = available_sources[data["source"]](data["config"])
    else:
        state.client.sources[data["source"]].add_to_config(data["config"])


@sio.on("config")
async def handle_config(sid: str, data: dict[str, Any]) -> None:
    """
    Handle the "config" message.

    This is called, when a source wants its configuration transmitted in
    a single message, rather than chunks. A source will be created with the
    given configuration.

    :param sid: The session id of the playback client
    :type sid: str
    :param data: A dictionary with a "source" (str) and a
        "config" (dict[str, Any]) entry. The exact content of the config entry
        depends on the source.
    :type data: dict[str, Any]
    :rtype: None
    """
    async with sio.session(sid) as session:
        room = session["room"]
    state = clients[room]

    if sid != state.sid:
        return

    state.client.sources[data["source"]] = available_sources[data["source"]](data["config"])


@sio.on("register-web")
async def handle_register_web(sid: str, data: dict[str, Any]) -> bool:
    """
    Handle a "register-web" message.

    Adds a web client to a requested room and sends it the initial state of the
    queue and recent list.

    :param sid: The session id of the web client.
    :type sid: str
    :param data: A dictionary, containing at least a "room" entry.
    :type data: dict[str, Any]
    :returns: True, if the room exist, False otherwise
    :rtype: bool
    """
    if data["room"] in clients:
        async with sio.session(sid) as session:
            session["room"] = data["room"]
            await sio.enter_room(sid, session["room"])
        state = clients[session["room"]]
        await send_state(state, sid)
        return True
    return False


@sio.on("register-admin")
async def handle_register_admin(sid: str, data: dict[str, Any]) -> bool:
    """
    Handle a "register-admin" message.

    If the client provides the correct secret for its room, the connection is
    upgraded to an admin connection.

    :param sid: The session id of the client, requesting admin.
    :type sid: str:
    :param data: A dictionary with at least a "secret" entry.
    :type data: dict[str, Any]
    :returns: True, if the secret is correct, False otherwise
    :rtype: bool
    """
    async with sio.session(sid) as session:
        room = session["room"]
    state = clients[room]

    is_admin: bool = data["secret"] == state.client.config["secret"]
    async with sio.session(sid) as session:
        session["admin"] = is_admin
    return is_admin


@sio.on("skip-current")
async def handle_skip_current(sid: str) -> None:
    """
    Handle a "skip-current" message.

    If this comes from an admin connection, forward the "skip-current" message
    to the playback client. This will be handled by the
    :py:func:`syng.client.handle_skip_current` function.

    :param sid: The session id of the client, requesting.
    :type sid: str
    :rtype: None
    """
    async with sio.session(sid) as session:
        room = session["room"]
        is_admin = session["admin"]
    state = clients[room]

    if is_admin:
        old_entry = await discard_first(room)
        await sio.emit("skip-current", old_entry, room=clients[room].sid)
        await send_state(state, room)


@sio.on("move-up")
async def handle_move_up(sid: str, data: dict[str, Any]) -> None:
    """
    Handle the "move-up" message.

    If on an admin connection, moves up the entry specified in the data by one
    place in the queue.

    :param sid: The session id of the client requesting.
    :type sid: str
    :param data: A dictionary with at least an "uuid" entry
    :type data: dict[str, Any]
    :rtype: None
    """
    async with sio.session(sid) as session:
        room = session["room"]
        is_admin = session["admin"]
    state = clients[room]
    if is_admin:
        await state.queue.move_up(data["uuid"])
        await send_state(state, room)


@sio.on("skip")
async def handle_skip(sid: str, data: dict[str, Any]) -> None:
    """
    Handle the "skip" message.

    If on an admin connection, removes the entry specified by data["uuid"]
    from the queue or the waiting room. Triggers the waiting room.

    :param sid: The session id of the client requesting.
    :type sid: str
    :param data: A dictionary with at least an "uuid" entry.
    :type data: dict[str, Any]
    :rtype: None
    """
    async with sio.session(sid) as session:
        room = session["room"]
        is_admin = session["admin"]
    state = clients[room]

    if is_admin:
        entry = state.queue.find_by_uuid(data["uuid"])
        if entry is not None:
            logger.info("Skipping %s", entry)

            await add_songs_from_waiting_room(room)

            await state.queue.remove(entry)

        first_entry_index = None
        for idx, wr_entry in enumerate(state.waiting_room):
            if wr_entry.uuid == data["uuid"]:
                first_entry_index = idx
                break

        if first_entry_index is not None:
            logger.info(
                "Deleting %s from waiting room",
                state.waiting_room[first_entry_index],
            )
            del state.waiting_room[first_entry_index]
        await send_state(state, room)


@sio.on("disconnect")
async def handle_disconnect(sid: str) -> None:
    """
    Handle the "disconnect" message.

    This message is send automatically, when a client disconnets.

    Remove the client from its room.

    :param sid: The session id of the client disconnecting
    :type sid: str
    :rtype: None
    """
    async with sio.session(sid) as session:
        if "room" in session:
            await sio.leave_room(sid, session["room"])


@sio.on("search")
async def handle_search(sid: str, data: dict[str, Any]) -> None:
    """
    Handle the "search" message.

    Forwards the dict["query"] to the :py:func:`Source.search` method, and
    execute them concurrently. The order is given by the
    :py:attr:`Config.sources_prio` attribute of the state.

    The result will be send with a "search-results" message to the (web)
    client.

    :param sid: The session id of the client requesting.
    :type sid: str
    :param data: A dictionary with at least a "query" entry.
    :type data: dict[str, str]
    :rtype: None
    """
    async with sio.session(sid) as session:
        room = session["room"]
    state = clients[room]

    query = data["query"]
    if (
        app["type"] != "restricted"
        or "key" in state.client.config
        and check_registration(state.client.config["key"])
    ):
        results_list = await asyncio.gather(
            *[state.client.sources[source].search(query) for source in state.client.sources_prio]
        )

        results = [
            search_result for source_result in results_list for search_result in source_result
        ]
        await send_search_results(sid, results)
    else:
        print("Denied")
        await sio.emit("search", {"query": query, "sid": sid}, room=state.sid)


@sio.on("search-results")
async def handle_search_results(sid: str, data: dict[str, Any]) -> None:
    """
    Handle the "search-results" message.

    This message is send by the playback client, once it has received search
    results. The results are send to the web client.

    The data dictionary should have the following keys:
        - `sid`, the session id of the web client (str)
        - `results`, a list of search results (list[dict[str, Any]])

    :param sid: The session id of the playback client
    :type sid: str
    :param data: A dictionary with the keys described above
    :type data: dict[str, Any]
    :rtype: None
    """
    async with sio.session(sid) as session:
        room = session["room"]
    state = clients[room]

    if sid != state.sid:
        return

    web_sid = data["sid"]
    results = [Result.from_dict(result) for result in data["results"]]

    # TODO: we convert the results to YouTube objects. This
    # adds them to the cache to prevent YouTube from blocking us.
    __unused_yt_list = [
        YouTube.from_result(
            {
                "duration": result.duration,
                "title": result.title,
                "channel": result.artist,
                "url": result.ident,
            }
        )
        for result in results
        if "youtube" == result.source
    ]

    await send_search_results(web_sid, results)


async def send_search_results(sid: str, results: list[Result]) -> None:
    """
    Send search results to a client.

    :param sid: The session id of the client to send the results to.
    :type sid: str
    :param results: The search results to send.
    :type results: list[Result]
    :rtype: None
    """
    await sio.emit(
        "search-results",
        {"results": results},
        room=sid,
    )


async def cleanup() -> None:
    """
    Clean up the unused playback clients

    This runs every hour, and removes every client, that did not requested a song for four hours.

    :rtype: None
    """

    logger.info("Start Cleanup")
    to_remove: list[str] = []
    for sid, state in clients.items():
        logger.info("Client %s, last seen: %s", sid, str(state.last_seen))
        if state.last_seen + datetime.timedelta(hours=4) < datetime.datetime.now():
            logger.info("No activity for 4 hours, removing %s", sid)
            to_remove.append(sid)
    for sid in to_remove:
        await sio.disconnect(sid)
        del clients[sid]
    logger.info("End Cleanup")

    # The internal loop counter does not use a regular timestamp, so we need to convert between
    # regular datetime and the async loop time
    now = datetime.datetime.now()
    # today = datetime.datetime(now.year, now.month, now.day)
    # next_run = today + datetime.timedelta(days=1)

    next_run = now + datetime.timedelta(hours=1)
    offset = next_run.timestamp() - now.timestamp()
    loop_next = asyncio.get_event_loop().time() + offset

    logger.info("Next Cleanup at %s", str(next))
    asyncio.get_event_loop().call_at(loop_next, lambda: asyncio.create_task(cleanup()))


async def background_tasks(
    iapp: web.Application,
) -> AsyncGenerator[None, None]:
    """
    Create all the background tasks.

    For now, this is only the cleanup task.

    :param iapp: The web application
    :type iapp: web.Application
    :rtype: AsyncGenerator[None, None]
    """

    iapp["repeated_cleanup"] = asyncio.create_task(cleanup())

    yield

    iapp["repeated_cleanup"].cancel()
    await iapp["repeated_cleanup"]


def run_server(args: Namespace) -> None:
    """
    Run the server.

    `args` consists of the following attributes:
        - `host`, the host to bind to
        - `port`, the port to bind to
        - `root_folder`, the root folder of the web client
        - `registration_keyfile`, the file containing the registration keys
        - `private`, if the server is private
        - `restricted`, if the server is restricted

    :param args: The command line arguments
    :type args: Namespace
    :rtype: None
    """
    app["type"] = "private" if args.private else "restricted" if args.restricted else "public"
    if args.registration_keyfile:
        app["registration-keyfile"] = args.registration_keyfile

    app["root_folder"] = args.root_folder

    app.add_routes([web.static("/assets/", os.path.join(app["root_folder"], "assets/"))])

    app.router.add_route("*", "/", root_handler)
    app.router.add_route("*", "/{room}", root_handler)
    app.router.add_route("*", "/{room}/", root_handler)

    app.cleanup_ctx.append(background_tasks)

    web.run_app(app, host=args.host, port=args.port)



from io import BytesIO
import logging
from logging.handlers import QueueListener
from multiprocessing import Process, Queue
from collections.abc import Callable
from datetime import datetime
import os
import builtins
from functools import partial
import random
from typing import Any, Optional
import multiprocessing
import secrets
import string
import signal

from PyQt6.QtCore import QTimer
from PyQt6.QtGui import QCloseEvent, QIcon, QPixmap
from PyQt6.QtWidgets import (
    QApplication,
    QCheckBox,
    QComboBox,
    QDateTimeEdit,
    QFormLayout,
    QHBoxLayout,
    QLabel,
    QLineEdit,
    QMainWindow,
    QPushButton,
    QSizePolicy,
    QSpacerItem,
    QTabWidget,
    QVBoxLayout,
    QWidget,
)
from yaml import dump, load, Loader, Dumper
from qrcode.main import QRCode
import platformdirs

from .client import create_async_and_start_client, default_config

from .sources import available_sources

# try:
#     from .server import run_server
#
#     SERVER_AVAILABLE = True
# except ImportError:
#     if TYPE_CHECKING:
#         from .server import run_server
#
#     SERVER_AVAILABLE = False


# TODO: ScrollableFrame
class OptionFrame(QWidget):
    def add_bool_option(self, name: str, description: str, value: bool = False) -> None:
        label = QLabel(description, self)

        self.bool_options[name] = QCheckBox(self)
        self.bool_options[name].setChecked(value)
        self.form_layout.addRow(label, self.bool_options[name])
        self.number_of_options += 1

    def add_string_option(
        self,
        name: str,
        description: str,
        value: Optional[str] = "",
        callback: Optional[Callable[..., None]] = None,
        is_password: bool = False,
    ) -> None:
        if value is None:
            value = ""

        label = QLabel(description, self)

        self.string_options[name] = QLineEdit(self)
        if is_password:
            self.string_options[name].setEchoMode(QLineEdit.EchoMode.Password)
            action = self.string_options[name].addAction(
                QIcon.fromTheme("dialog-password"), QLineEdit.ActionPosition.TrailingPosition
            )
            if action is not None:
                action.triggered.connect(
                    lambda: self.string_options[name].setEchoMode(
                        QLineEdit.EchoMode.Normal
                        if self.string_options[name].echoMode() == QLineEdit.EchoMode.Password
                        else QLineEdit.EchoMode.Password
                    )
                )

        self.string_options[name].insert(value)
        self.form_layout.addRow(label, self.string_options[name])
        if callback is not None:
            self.string_options[name].textChanged.connect(callback)
        self.number_of_options += 1

    def del_list_element(
        self,
        name: str,
        element: QLineEdit,
        line: QWidget,
        layout: QVBoxLayout,
    ) -> None:
        self.list_options[name].remove(element)

        layout.removeWidget(line)
        line.deleteLater()

    def add_list_element(
        self,
        name: str,
        layout: QVBoxLayout,
        init: str,
        callback: Optional[Callable[..., None]],
    ) -> None:
        input_and_minus = QWidget()
        input_and_minus_layout = QHBoxLayout(input_and_minus)
        input_and_minus.setLayout(input_and_minus_layout)

        input_and_minus_layout.setContentsMargins(0, 0, 0, 0)

        input_field = QLineEdit(input_and_minus)
        input_field.setText(init)
        input_field.setSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Fixed)
        input_and_minus_layout.addWidget(input_field)
        if callback is not None:
            input_field.textChanged.connect(callback)

        minus_button = QPushButton(QIcon.fromTheme("list-remove"), "", input_and_minus)
        minus_button.clicked.connect(
            partial(self.del_list_element, name, input_field, input_and_minus, layout)
        )
        minus_button.setFixedWidth(40)
        minus_button.setSizePolicy(QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Fixed)
        input_and_minus_layout.addWidget(minus_button)

        layout.insertWidget(layout.count() - 1, input_and_minus)

        self.list_options[name].append(input_field)

    def add_list_option(
        self,
        name: str,
        description: str,
        value: list[str],
        callback: Optional[Callable[..., None]] = None,
    ) -> None:
        label = QLabel(description, self)

        container_layout = QVBoxLayout()

        self.form_layout.addRow(label, container_layout)

        self.list_options[name] = []
        for v in value:
            self.add_list_element(name, container_layout, v, callback)
        plus_button = QPushButton(QIcon.fromTheme("list-add"), "", self)
        plus_button.clicked.connect(
            partial(self.add_list_element, name, container_layout, "", callback)
        )
        plus_button.setSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Fixed)

        container_layout.addWidget(plus_button)

        self.number_of_options += 1

    def add_choose_option(
        self, name: str, description: str, values: list[str], value: str = ""
    ) -> None:
        label = QLabel(description, self)

        self.choose_options[name] = QComboBox(self)
        self.choose_options[name].addItems(values)
        self.choose_options[name].setCurrentText(value)
        self.form_layout.addRow(label, self.choose_options[name])
        self.number_of_options += 1

    def add_date_time_option(self, name: str, description: str, value: str) -> None:
        label = QLabel(description, self)
        date_time_layout = QHBoxLayout()
        date_time_widget = QDateTimeEdit(self)
        date_time_enabled = QCheckBox("Enabled", self)
        date_time_enabled.stateChanged.connect(
            lambda: date_time_widget.setEnabled(date_time_enabled.isChecked())
        )

        self.date_time_options[name] = (date_time_widget, date_time_enabled)
        date_time_widget.setCalendarPopup(True)
        try:
            date_time_widget.setDateTime(datetime.fromisoformat(value))
            date_time_enabled.setChecked(True)
        except (TypeError, ValueError):
            date_time_widget.setDateTime(datetime.now())
            date_time_widget.setEnabled(False)
            date_time_enabled.setChecked(False)

        date_time_layout.addWidget(date_time_widget)
        date_time_layout.addWidget(date_time_enabled)

        self.form_layout.addRow(label, date_time_layout)

        self.number_of_options += 1

    def __init__(self, parent: Optional[QWidget] = None) -> None:
        super().__init__(parent)
        self.form_layout = QFormLayout(self)
        self.setLayout(self.form_layout)
        self.number_of_options: int = 0
        self.string_options: dict[str, QLineEdit] = {}
        self.choose_options: dict[str, QComboBox] = {}
        self.bool_options: dict[str, QCheckBox] = {}
        self.list_options: dict[str, list[QLineEdit]] = {}
        self.date_time_options: dict[str, tuple[QDateTimeEdit, QCheckBox]] = {}

    def get_config(self) -> dict[str, Any]:
        config: dict[str, Any] = {}
        for name, textbox in self.string_options.items():
            config[name] = textbox.text().strip()

        for name, optionmenu in self.choose_options.items():
            config[name] = optionmenu.currentText().strip()

        for name, checkbox in self.bool_options.items():
            config[name] = checkbox.isChecked()

        for name, textboxes in self.list_options.items():
            config[name] = []
            for textbox in textboxes:
                config[name].append(textbox.text().strip())

        for name, (picker, checkbox) in self.date_time_options.items():
            if not checkbox.isChecked():
                config[name] = None
                continue
            try:
                config[name] = picker.dateTime().toPyDateTime().isoformat()
            except ValueError:
                config[name] = None

        return config


class SourceTab(OptionFrame):
    def __init__(self, parent: QWidget, source_name: str, config: dict[str, Any]) -> None:
        super().__init__(parent)
        source = available_sources[source_name]
        self.vars: dict[str, str | bool | list[str]] = {}
        for name, (typ, desc, default) in source.config_schema.items():
            value = config[name] if name in config else default
            match typ:
                case builtins.bool:
                    self.add_bool_option(name, desc, value=value)
                case builtins.list:
                    self.add_list_option(name, desc, value=value)
                case builtins.str:
                    self.add_string_option(name, desc, value=value)
                case "password":
                    self.add_string_option(name, desc, value=value, is_password=True)


class GeneralConfig(OptionFrame):
    def __init__(
        self,
        parent: QWidget,
        config: dict[str, Any],
        callback: Callable[..., None],
    ) -> None:
        super().__init__(parent)

        self.add_string_option("server", "Server", config["server"], callback)
        self.add_string_option("room", "Room", config["room"], callback)
        self.add_string_option("secret", "Admin Password", config["secret"], is_password=True)
        self.add_choose_option(
            "waiting_room_policy",
            "Waiting room policy",
            ["forced", "optional", "none"],
            str(config["waiting_room_policy"]).lower(),
        )
        self.add_date_time_option("last_song", "Last song ends at", config["last_song"])
        self.add_string_option(
            "preview_duration", "Preview duration in seconds", str(config["preview_duration"])
        )
        self.add_string_option(
            "key", "Key for server (if necessary)", config["key"], is_password=True
        )

    def get_config(self) -> dict[str, Any]:
        config = super().get_config()
        try:
            config["preview_duration"] = int(config["preview_duration"])
        except ValueError:
            config["preview_duration"] = 0

        return config


class SyngGui(QMainWindow):
    def closeEvent(self, a0: Optional[QCloseEvent]) -> None:
        if self.syng_server is not None:
            self.syng_server.kill()
            self.syng_server.join()

        if self.syng_client is not None:
            self.syng_client.terminate()
            self.syng_client.join(1.0)
            self.syng_client.kill()

        self.destroy()

    def add_buttons(self) -> None:
        self.buttons_layout = QHBoxLayout()
        self.central_layout.addLayout(self.buttons_layout)

        # self.startsyng_serverbutton = QPushButton("Start Local Server")
        # self.startsyng_serverbutton.clicked.connect(self.start_syng_server)
        # self.buttons_layout.addWidget(self.startsyng_serverbutton)

        spacer_item = QSpacerItem(40, 20, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)
        self.notification_label = QLabel("", self)
        spacer_item2 = QSpacerItem(40, 20, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)
        self.buttons_layout.addItem(spacer_item)
        self.buttons_layout.addWidget(self.notification_label)
        self.buttons_layout.addItem(spacer_item2)

        self.savebutton = QPushButton("Apply")
        self.savebutton.clicked.connect(self.save_config)
        self.buttons_layout.addWidget(self.savebutton)

        self.startbutton = QPushButton("Apply and Start")

        self.startbutton.clicked.connect(self.start_syng_client)
        self.buttons_layout.addWidget(self.startbutton)

    def init_frame(self) -> None:
        self.frm = QHBoxLayout()
        self.central_layout.addLayout(self.frm)

    def init_tabs(self) -> None:
        self.tabview = QTabWidget(parent=self.central_widget)
        self.tabview.setAcceptDrops(False)
        self.tabview.setTabPosition(QTabWidget.TabPosition.West)
        self.tabview.setTabShape(QTabWidget.TabShape.Rounded)
        self.tabview.setDocumentMode(False)
        self.tabview.setTabsClosable(False)
        self.tabview.setObjectName("tabWidget")

        self.tabview.setTabText(0, "General")
        for i, source in enumerate(available_sources):
            self.tabview.setTabText(i + 1, source)

        self.frm.addWidget(self.tabview)

    def add_qr(self) -> None:
        self.qr_widget = QWidget(parent=self.central_widget)
        self.qr_layout = QVBoxLayout(self.qr_widget)
        self.qr_widget.setLayout(self.qr_layout)

        self.qr_label = QLabel(self.qr_widget)
        self.linklabel = QLabel(self.qr_widget)

        self.qr_layout.addWidget(self.qr_label)
        self.qr_layout.addWidget(self.linklabel)

        self.linklabel.setOpenExternalLinks(True)

        self.frm.addWidget(self.qr_widget)

    def add_general_config(self, config: dict[str, Any]) -> None:
        self.general_config = GeneralConfig(self, config, self.update_qr)
        self.tabview.addTab(self.general_config, "General")

    def add_source_config(self, source_name: str, source_config: dict[str, Any]) -> None:
        self.tabs[source_name] = SourceTab(self, source_name, source_config)
        self.tabview.addTab(self.tabs[source_name], source_name)

    def __init__(self) -> None:
        super().__init__()
        self.setWindowTitle("Syng")

        rel_path = os.path.dirname(__file__)
        qt_img = QPixmap(os.path.join(rel_path, "static/syng.png"))
        self.qt_icon = QIcon(qt_img)
        self.setWindowIcon(self.qt_icon)

        self.syng_server: Optional[Process] = None
        # self.syng_client: Optional[subprocess.Popen[bytes]] = None
        self.syng_client: Optional[Process] = None
        self.syng_client_logging_listener: Optional[QueueListener] = None

        self.configfile = os.path.join(platformdirs.user_config_dir("syng"), "config.yaml")

        try:
            with open(self.configfile, encoding="utf8") as cfile:
                loaded_config = load(cfile, Loader=Loader)
        except FileNotFoundError:
            print("No config found, using default values")
            loaded_config = {}
        config: dict[str, dict[str, Any]] = {"sources": {}, "config": default_config()}

        try:
            config["config"] |= loaded_config["config"]
        except (KeyError, TypeError):
            print("Could not load config")

        if not config["config"]["secret"]:
            config["config"]["secret"] = "".join(
                secrets.choice(string.ascii_letters + string.digits) for _ in range(8)
            )

        if config["config"]["room"] == "":
            config["config"]["room"] = "".join(
                [random.choice(string.ascii_letters) for _ in range(6)]
            ).upper()

        self.central_widget = QWidget(parent=self)
        self.central_layout = QVBoxLayout(self.central_widget)

        self.init_frame()
        self.init_tabs()
        self.add_buttons()
        self.add_qr()
        self.add_general_config(config["config"])
        self.tabs: dict[str, SourceTab] = {}

        for source_name in available_sources:
            try:
                source_config = loaded_config["sources"][source_name]
            except (KeyError, TypeError):
                source_config = {}

            self.add_source_config(source_name, source_config)

        self.update_qr()

        self.setCentralWidget(self.central_widget)

        # check every 500 ms if client is running
        self.timer = QTimer()
        self.timer.timeout.connect(self.check_if_client_is_running)

    def save_config(self) -> None:
        os.makedirs(os.path.dirname(self.configfile), exist_ok=True)

        with open(self.configfile, "w", encoding="utf-8") as f:
            dump(self.gather_config(), f, Dumper=Dumper)

    def gather_config(self) -> dict[str, Any]:
        sources = {}
        for source, tab in self.tabs.items():
            sources[source] = tab.get_config()

        general_config = self.general_config.get_config()

        return {"sources": sources, "config": general_config}

    def check_if_client_is_running(self) -> None:
        if self.syng_client is None:
            self.timer.stop()
            return

        if not self.syng_client.is_alive():
            self.syng_client = None
            self.set_client_button_start()
        else:
            self.set_client_button_stop()

    def set_client_button_stop(self) -> None:
        self.startbutton.setText("Stop")

    def set_client_button_start(self) -> None:
        self.startbutton.setText("Save and Start")

    def start_syng_client(self) -> None:
        if self.syng_client is None or not self.syng_client.is_alive():
            self.save_config()
            config = self.gather_config()
            queue: Queue[logging.LogRecord] = multiprocessing.Queue()

            self.syng_client_logging_listener = QueueListener(
                queue, LoggingLabelHandler(self.notification_label)
            )
            self.syng_client_logging_listener.start()

            self.syng_client = multiprocessing.Process(
                target=create_async_and_start_client, args=[config, queue]
            )
            self.syng_client.start()
            self.notification_label.setText("")
            self.timer.start(500)
            self.set_client_button_stop()
        else:
            self.syng_client.terminate()
            self.syng_client.join(1.0)
            self.syng_client.kill()
            self.set_client_button_start()

    # def start_syng_server(self) -> None:
    #     if self.syng_server is None:
    #         root_path = os.path.join(os.path.dirname(__file__), "static")
    #         self.syng_server = multiprocessing.Process(
    #             target=run_server,
    #             args=[
    #                 Namespace(
    #                     host="0.0.0.0",
    #                     port=8080,
    #                     registration_keyfile=None,
    #                     root_folder=root_path,
    #                     private=False,
    #                     restricted=False,
    #                 )
    #             ],
    #         )
    #         self.syng_server.start()
    #         self.startsyng_serverbutton.setText("Stop Local Server")
    #     else:
    #         self.syng_server.terminate()
    #         self.syng_server.join()
    #         self.syng_server = None
    #         self.startsyng_serverbutton.setText("Start Local Server")

    def change_qr(self, data: str) -> None:
        qr = QRCode(box_size=10, border=2)
        qr.add_data(data)
        qr.make()
        image = qr.make_image().convert("RGB")
        buf = BytesIO()
        image.save(buf, "PNG")
        qr_pixmap = QPixmap()
        qr_pixmap.loadFromData(buf.getvalue(), "PNG")
        self.qr_label.setPixmap(qr_pixmap)

    def update_qr(self) -> None:
        config = self.general_config.get_config()
        syng_server = config["server"]
        syng_server += "" if syng_server.endswith("/") else "/"
        room = config["room"]
        self.linklabel.setText(
            f'<center><a href="{syng_server + room}">{syng_server + room}</a><center>'
        )
        self.linklabel.setSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Fixed)
        self.change_qr(syng_server + room)


class LoggingLabelHandler(logging.Handler):
    def __init__(self, label: QLabel):
        super().__init__()
        self.label = label

    def emit(self, record: logging.LogRecord) -> None:
        self.label.setText(self.format(record))


def run_gui() -> None:
    signal.signal(signal.SIGINT, signal.SIG_DFL)

    app = QApplication([])
    app.setApplicationName("Syng")
    app.setDesktopFileName("rocks.syng.Syng")
    window = SyngGui()
    window.show()
    app.exec()



"""
Module for the playback client.

The client connects to the server via the socket.io protocol, and plays the
songs, that are sent by the server.

Playback is done by the :py:class:`syng.sources.source.Source` objects, that
are configured in the `sources` section of the configuration file and can currently
be one of:
  - `youtube`
  - `s3`
  - `files`
"""

from __future__ import annotations
import asyncio
import datetime
import logging
from logging.handlers import QueueHandler
from multiprocessing import Queue
import secrets
import string
import tempfile
import signal
from argparse import Namespace
from dataclasses import dataclass
from dataclasses import field
from traceback import print_exc
from typing import Any, Optional

from qrcode.main import QRCode

import socketio
import engineio
from PIL import Image
from yaml import load, Loader

from . import jsonencoder
from .entry import Entry
from .sources import configure_sources, Source


sio: socketio.AsyncClient = socketio.AsyncClient(json=jsonencoder)
logger: logging.Logger = logging.getLogger(__name__)
sources: dict[str, Source] = {}


currentLock: asyncio.Semaphore = asyncio.Semaphore(0)


def default_config() -> dict[str, Optional[int | str]]:
    """
    Return a default configuration for the client.

    :returns: A dictionary with the default configuration.
    :rtype: dict[str, Optional[int | str]]
    """
    return {
        "server": "https://syng.rocks",
        "room": "",
        "preview_duration": 3,
        "secret": None,
        "last_song": None,
        "waiting_room_policy": None,
        "key": None,
    }


@dataclass
class State:
    """This captures the current state of the playback client.

    It doubles as a backup of the state of the :py:class:`syng.server` in case
    the server needs to be restarted.

    :param current_source: This holds a reference to the
        :py:class:`syng.sources.source.Source` object, that is currently
        playing. If no song is played, the value is `None`.
    :type current_source: Optional[Source]
    :param queue: A copy of the current playlist on the server.
    :type queue: list[Entry]
    :param waiting_room: A copy of the waiting room on the server.
    :type waiting_room: list[Entry]
    :param recent: A copy of all played songs this session.
    :type recent: list[Entry]
    :param config: Various configuration options for the client:
        * `server` (`str`): The url of the server to connect to.
        * `room` (`str`): The room on the server this playback client is connected to.
        * `secret` (`str`): The passcode of the room. If a playback client reconnects to
            a room, this must be identical. Also, if a webclient wants to have
            admin privileges, this must be included.
        * `key` (`Optional[str]`) An optional key, if registration or functionality on the server
            is limited.
        * `preview_duration` (`Optional[int]`): The duration in seconds the
            playback client shows a preview for the next song. This is accounted for
            in the calculation of the ETA for songs later in the queue.
        * `last_song` (`Optional[datetime.datetime]`): A timestamp, defining the end of
            the queue.
        * `waiting_room_policy` (Optional[str]): One of:
            - `forced`, if a performer is already in the queue, they are put in the
                       waiting room.
            - `optional`, if a performer is already in the queue, they have the option
                          to be put in the waiting room.
            - `None`, performers are always added to the queue.
    :type config: dict[str, Any]:
    """

    # pylint: disable=too-many-instance-attributes

    current_source: Optional[Source] = None
    queue: list[Entry] = field(default_factory=list)
    waiting_room: list[Entry] = field(default_factory=list)
    recent: list[Entry] = field(default_factory=list)
    config: dict[str, Any] = field(default_factory=default_config)


state: State = State()


@sio.on("update_config")
async def handle_update_config(data: dict[str, Any]) -> None:
    """
    Handle the "update_config" message.

    Currently, this function is untested and should be considered dangerous.

    :param data: A dictionary with the new configuration.
    :type data: dict[str, Any]
    :rtype: None
    """
    state.config = default_config() | data


@sio.on("skip-current")
async def handle_skip_current(data: dict[str, Any]) -> None:
    """
    Handle the "skip-current" message.

    Skips the song, that is currently played. If playback currently waits for
    buffering, the buffering is also aborted.

    Since the ``queue`` could already be updated, when this evaluates, the
    first entry in the queue is send explicitly.

    :param data: An entry, that should be equivalent to the first entry of the
        queue.
    :rtype: None
    """
    logger.info("Skipping current")
    if state.current_source is not None:
        await state.current_source.skip_current(Entry(**data))


@sio.on("state")
async def handle_state(data: dict[str, Any]) -> None:
    """
    Handle the "state" message.

    The "state" message forwards the current queue and recent list from the
    server. This function saves a copy of both in the global
    :py:class:`State`:.

    After recieving the new state, a buffering task for the first elements of
    the queue is started.

    :param data: A dictionary with the `queue` and `recent` list.
    :type data: dict[str, Any]
    :rtype: None
    """
    state.queue = [Entry(**entry) for entry in data["queue"]]
    state.waiting_room = [Entry(**entry) for entry in data["waiting_room"]]
    state.recent = [Entry(**entry) for entry in data["recent"]]

    for entry in state.queue[:2]:
        logger.info("Buffering: %s", entry.title)
        await sources[entry.source].buffer(entry)


@sio.on("connect")
async def handle_connect() -> None:
    """
    Handle the "connect" message.

    Called when the client successfully connects or reconnects to the server.
    Sends a `register-client` message to the server with the initial state and
    configuration of the client, consiting of the currently saved
    :py:attr:`State.queue` and :py:attr:`State.recent` field of the global
    :py:class:`State`, as well a room code the client wants to connect to, a
    secret to secure the access to the room and a config dictionary.

    If the room code is `None`, the server will issue a room code.

    This message will be handled by the
    :py:func:`syng.server.handle_register_client` function of the server.

    :rtype: None
    """
    logger.info("Connected to server")
    data = {
        "queue": state.queue,
        "waiting_room": state.waiting_room,
        "recent": state.recent,
        "config": state.config,
    }
    await sio.emit("register-client", data)


@sio.on("get-meta-info")
async def handle_get_meta_info(data: dict[str, Any]) -> None:
    """
    Handle a "get-meta-info" message.

    Collects the metadata for a given :py:class:`Entry`, from its source, and
    sends them back to the server in a "meta-info" message. On the server side
    a :py:func:`syng.server.handle_meta_info` function is called.

    :param data: A dictionary encoding the entry
    :type data: dict[str, Any]
    :rtype: None
    """
    source: Source = sources[data["source"]]
    meta_info: dict[str, Any] = await source.get_missing_metadata(Entry(**data))
    await sio.emit("meta-info", {"uuid": data["uuid"], "meta": meta_info})


async def preview(entry: Entry) -> None:
    """
    Generate and play a preview for a given :py:class:`Entry`.

    This function shows a black screen and prints the artist, title and
    performer of the entry for a duration.

    This is done by creating a black png file, and showing subtitles in the
    middle of the screen.... don't ask, it works

    :param entry: The entry to preview
    :type entry: :py:class:`Entry`
    :rtype: None
    """
    background = Image.new("RGB", (1280, 720))
    subtitle: str = f"""1
00:00:00,00 --> 00:05:00,00
{entry.artist} - {entry.title}
{entry.performer}"""
    with tempfile.NamedTemporaryFile() as tmpfile:
        background.save(tmpfile, "png")
        process = await asyncio.create_subprocess_exec(
            "mpv",
            tmpfile.name,
            f"--image-display-duration={state.config['preview_duration']}",
            "--sub-pos=50",
            "--sub-file=-",
            "--fullscreen",
            stdin=asyncio.subprocess.PIPE,
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.PIPE,
        )
        await process.communicate(subtitle.encode())


@sio.on("play")
async def handle_play(data: dict[str, Any]) -> None:
    """
    Handle the "play" message.

    Plays the :py:class:`Entry`, that is encoded in the `data` parameter. If a
    :py:attr:`State.preview_duration` is set, it shows a small preview before
    that.

    When the playback is done, the next song is requested from the server with
    a "pop-then-get-next" message. This is handled by the
    :py:func:`syng.server.handle_pop_then_get_next` function on the server.

    If the entry is marked as skipped, emit a "get-first"  message instead,
    because the server already handled the removal of the first entry.

    :param data: A dictionary encoding the entry
    :type data: dict[str, Any]
    :rtype: None
    """
    entry: Entry = Entry(**data)
    print(
        f"Playing: {entry.artist} - {entry.title} [{entry.album}] "
        f"({entry.source}) for {entry.performer}"
    )
    try:
        state.current_source = sources[entry.source]
        if state.config["preview_duration"] > 0:
            await preview(entry)
        await sources[entry.source].play(entry)
    except Exception:  # pylint: disable=broad-except
        print_exc()
    state.current_source = None
    if entry.skip:
        await sio.emit("get-first")
    else:
        await sio.emit("pop-then-get-next")


@sio.on("search")
async def handle_search(data: dict[str, Any]) -> None:
    """
    Handle the "search" message.

    This handles client side search requests. It sends a search request to all
    configured :py:class:`syng.sources.source.Source` and collects the results.

    The results are then send back to the server in a "search-results" message,
    including the `sid` of the corresponding webclient.

    :param data: A dictionary with the `query` and `sid` entry.
    :type data: dict[str, Any]
    :rtype: None
    """
    logger.info(f"Searching for: {data['query']}")
    query = data["query"]
    sid = data["sid"]
    results_list = await asyncio.gather(*[source.search(query) for source in sources.values()])

    results = [
        search_result.to_dict() for source_result in results_list for search_result in source_result
    ]

    await sio.emit("search-results", {"results": results, "sid": sid})


@sio.on("client-registered")
async def handle_client_registered(data: dict[str, Any]) -> None:
    """
    Handle the "client-registered" message.

    If the registration was successfull (`data["success"]` == `True`), store
    the room code in the global :py:class:`State` and print out a link to join
    the webclient.

    Start listing all configured :py:class:`syng.sources.source.Source` to the
    server via a "sources" message. This message will be handled by the
    :py:func:`syng.server.handle_sources` function and may request additional
    configuration for each source.

    If there is no song playing, start requesting the first song of the queue
    with a "get-first" message. This will be handled on the server by the
    :py:func:`syng.server.handle_get_first` function.

    :param data: A dictionary containing a `success` and a `room` entry.
    :type data: dict[str, Any]
    :rtype: None
    """
    if data["success"]:
        logger.info("Registered")
        print(f"Join here: {state.config['server']}/{data['room']}")
        qr = QRCode(box_size=20, border=2)
        qr.add_data(f"{state.config['server']}/{data['room']}")
        qr.make()
        qr.print_ascii()
        state.config["room"] = data["room"]
        await sio.emit("sources", {"sources": list(sources.keys())})
        if state.current_source is None:  # A possible race condition can occur here
            await sio.emit("get-first")
    else:
        logger.warning("Registration failed")
        await sio.disconnect()


@sio.on("request-config")
async def handle_request_config(data: dict[str, Any]) -> None:
    """
    Handle the "request-config" message.

    Sends the specific server side configuration for a given
    :py:class:`syng.sources.source.Source`.

    A Source can decide, that the config will be split up in multiple Parts.
    If this is the case, multiple "config-chunk" messages will be send with a
    running enumerator. Otherwise a singe "config" message will be send.

    :param data: A dictionary with the entry `source` and a string, that
        corresponds to the name of a source.
    :type data: dict[str, Any]
    :rtype: None
    """
    if data["source"] in sources:
        config: dict[str, Any] | list[dict[str, Any]] = await sources[data["source"]].get_config()
        if isinstance(config, list):
            num_chunks: int = len(config)
            for current, chunk in enumerate(config):
                await sio.emit(
                    "config-chunk",
                    {
                        "source": data["source"],
                        "config": chunk,
                        "number": current + 1,
                        "total": num_chunks,
                    },
                )
        else:
            await sio.emit("config", {"source": data["source"], "config": config})


def signal_handler() -> None:
    """
    Signal handler for the client.

    This function is called when the client receives a signal to terminate. It
    will disconnect from the server and kill the current player.

    :rtype: None
    """
    engineio.async_client.async_signal_handler()
    if state.current_source is not None:
        if state.current_source.player is not None:
            state.current_source.player.kill()


async def start_client(config: dict[str, Any]) -> None:
    """
    Initialize the client and connect to the server.

    :param config: Config options for the client
    :type config: dict[str, Any]
    :rtype: None
    """

    sources.update(configure_sources(config["sources"]))

    if "config" in config:
        last_song = (
            datetime.datetime.fromisoformat(config["config"]["last_song"]).timestamp()
            if "last_song" in config["config"] and config["config"]["last_song"]
            else None
        )
        state.config |= config["config"] | {"last_song": last_song}

    if not ("secret" in state.config and state.config["secret"]):
        state.config["secret"] = "".join(
            secrets.choice(string.ascii_letters + string.digits) for _ in range(8)
        )
        print(f"Generated secret: {state.config['secret']}")

    if not ("key" in state.config and state.config["key"]):
        state.config["key"] = ""

    await sio.connect(state.config["server"])

    asyncio.get_event_loop().add_signal_handler(signal.SIGINT, signal_handler)
    asyncio.get_event_loop().add_signal_handler(signal.SIGTERM, signal_handler)

    try:
        await sio.wait()
    except asyncio.CancelledError:
        pass
    finally:
        if state.current_source is not None:
            if state.current_source.player is not None:
                state.current_source.player.kill()


def create_async_and_start_client(
    config: dict[str, Any], queue: Optional[Queue[logging.LogRecord]] = None
) -> None:
    """
    Create an asyncio event loop and start the client.

    If a multiprocessing queue is given, the client will log to the queue.

    :param config: Config options for the client
    :type config: dict[str, Any]
    :param queue: A multiprocessing queue to log to
    :type queue: Optional[Queue[LogRecord]]
    :rtype: None
    """

    if queue is not None:
        logger.addHandler(QueueHandler(queue))

    asyncio.run(start_client(config))


def run_client(args: Namespace) -> None:
    """
    Run the client with the given arguments.

    Namespace contains the following attributes:
        - room: The room code to connect to
        - secret: The secret to connect to the room
        - config_file: The path to the configuration file
        - key: The key to connect to the server
        - server: The url of the server to connect to

    :param args: The arguments from the command line
    :type args: Namespace
    :rtype: None
    """
    try:
        with open(args.config_file, encoding="utf8") as file:
            config = load(file, Loader=Loader)
    except FileNotFoundError:
        config = {}

    if "config" not in config:
        config["config"] = {}

    if args.room:
        config["config"] |= {"room": args.room}
    if args.secret:
        config["config"] |= {"secret": args.secret}
    if args.server:
        config["config"] |= {"server": args.server}

    create_async_and_start_client(config)
